%
%    CLUSTER SEQUENTS
%
%
%    Refutability of cluster sequents is reduced to classical
%    satisfiability
%
%   To prove classica satisfiability we use logic2cnf SAT solver



:- module(cluster, [
		    is_ref_cluster_seq/2  % is_ref_cluster_seq(+Cluster_seq, -Result (true ore false),
		]
	 ).




:-use_module(prover).




%   Cluster_sequent := (GBox,GDiam, Gcl)  '=clu=>' Dcl
%
%     where
%     o GBox is a list of formulas box A  (A classical)
%     o GDiam is a list of formulas box diam B  (B classical)
%     o Gcl and Dcl are sets of classical formulas


%
%  is_ref_cluster_seq(+Cluster_seq, -Result)
%
%
%  Result = true  IF Cluster_seq is a refutable cluster sequent
%	    false otherwise
%



is_ref_cluster_seq(  (GBox,GDiam, Gcl)  '=clu=>' Dcl    , Result):-
	( GBox = [] ->  NegA_list = []
	 ;
	 setof( ~A , member(box A , GBox), NegA_list)
	),
	( Gcl = [] ->  NegG_list = []
	 ;
	 setof( ~G , member(G , Gcl), NegG_list)
	),
	union(NegA_list, NegG_list, Neg_list),
	union(Neg_list,Dcl, Z0),
	is_non_cl_valid_clause(Z0, R0), !,
	(
           R0 =  false ->
	      Result = false
	;
	  are_GDiam_non_cl_valid_clauses(GDiam, NegA_list, Result)
	).

%  are_GDiam_non_valid_clauses(GDiam, NegA_list, Result)
%
%  Let GDiam = [ diam B1, ... , diam Bm ]
%  and NegA_list = [~A1, ..., ~Al]
%
%  Result = true IF  all the clauses [~A1, ..., ~Al, ~Bi]  i = 1..m
%                    are not classically valid
%
%            false otherwise
%


 are_GDiam_non_cl_valid_clauses([diam B | GDiam ], NegA_list, Result) :-
	is_non_cl_valid_clause( [~B | NegA_list]   , Res), !,
	(
	   Res = false ->
	     Result = false
	;
	  are_GDiam_non_cl_valid_clauses( GDiam, NegA_list, Result)
	).


 are_GDiam_non_cl_valid_clauses([], _NegA_list, true).

% is_non_cl_valid(+Clause, -Result)
%
% Let Clause = [ F1, .... , Fn] ( n>=1 )
%
%  Result = true    IF the clause ~(F1 v ... v Fn)  cl-sat
%           false otherwise
%
%  If Clause = [], Result = true
%
%  NOTE:  F is not cl-valid IFF  ~F is cl-sat


is_non_cl_valid_clause(Clause, Result) :-
   Clause = [] ->
      Result = true
      ;
      (
       big_or(Clause, F),
       is_cl_sat_formula(~F,Result)
      ).





% big_and(+Formula_list : list of formulas, -F : formula)
%
% F = /\ Formula_list

big_and([A],A) :- !.

big_and([A | L], A & F1) :-
	big_and(L,F1).

% big_or(+Formula_list : list of formulas , -F : formula)
%
% F = \/ Formula_list

big_or([A],A) :- !.

big_or([A | L], A v F1) :-
	big_or(L,F1).




%  is_cl_sat_formula(+Form, -Result)
%
%  Result = true   IF  Form is a cl-sat formula
%           false  otherwise
%


is_cl_sat_formula(Form,Result) :-
	form_to_string(Form, StrForm,[],Atoms_in_Form),
	atomList_to_atomString(Atoms_in_Form,'',Str_Atoms_in_Form),
	atomic_list_concat(['def ', Str_Atoms_in_Form], Atoms_def),
	atomic_list_concat([Atoms_def, '; ',  StrForm, ';'], Input),
       	atomic_list_concat(['echo \" ',   Input, ' \"', ' | logic2cnf -j1 -p  | grep 0  >/dev/null   2> /dev/null '],Cmd),
	% if there is a solution, 0 is printed on stdout and Cmd succedes
	shell(Cmd,ExitStatus), !,
	(
	  ExitStatus == 0 ->  % Form is satisfiable
	     Result =  true
	    ;
	    Result =   false
	).


%%    form_to_string(+Form, -StringForm, +InAtomList, -OutAtomList)
%
%    StringForm is the string representing formula Form
%    according to logic2cnf syntax
%
%    OutAtomList = InAtomList U { a | a is an atom occurring in Form }
%


form_to_string(At,At,InAtoms, OutAtoms):-
	atom(At),!, union(InAtoms,[At],OutAtoms).

form_to_string(~F , Str ,InAtoms, OutAtoms):- !,
	form_to_string(F, S, InAtoms,OutAtoms),
	atomic_list_concat([ '~',  '(', S,  ')'],  Str).

form_to_string(F1 & F2, Str ,InAtoms, OutAtoms):- !,
	form_to_string(F1, S1, InAtoms,Out1),
	form_to_string(F2, S2, Out1,OutAtoms),
	atomic_list_concat([ '(', S1,  ')', '.', '(',  S2, ')'],  Str).

form_to_string(F1 v F2, Str ,InAtoms, OutAtoms):- !,
	form_to_string(F1, S1, InAtoms,Out1),
	form_to_string(F2, S2, Out1,OutAtoms),
	atomic_list_concat(['(', S1, ')', '+', '(', S2, ')'],  Str).

form_to_string(F1 --> F2, Str ,InAtoms, OutAtoms):- !,
	form_to_string(F1, S1, InAtoms,Out1),
	form_to_string(F2, S2, Out1,OutAtoms),
	atomic_list_concat(['(', S1, ')', '=>',  '(', S2, ')'],  Str).

form_to_string(F1 <--> F2, Str ,InAtoms, OutAtoms):- !,
	form_to_string(F1, S1, InAtoms,Out1),
	form_to_string(F2, S2, Out1,OutAtoms),
	atomic_list_concat(['(', S1, ')', '<=>',  '(', S2, ')'],  Str).


% atomsList_to_atomsString(+AtomList ,+In_str, +Out_string).
%
%  Out_string = In_str + string of Atom List, separated by ' '

atomList_to_atomString([],String,String).

atomList_to_atomString([At | L], In_str, Out_str):-
        atomList_to_atomString(L, In_str,Str1),
	atomic_list_concat( [At,' ', Str1], Out_str).














