%%   PROVER
%

:- module(prover, [
		   prove/2, % +Input ( Gamma==>Delta  ), -Proof_or_Refutation
		   is_atomic_formula/1,
		   is_classical_formula/1,
		   is_proof/1,
		   is_ref/1,
		   % OPERATORS
		   op(100, fy, 'box'),
		   op(100, fy, 'diam'),
		   op(100, fy, 'boxp'),  %  extended language
		   op(100, fy, 'diamc'), %  extended language
		   op(100, fy, ~),
		   op(150, xfy, &),
		   op(200, xfy, v),
		   op(300, xfy, '-->'),
		   op(250, xfy, '<-->'),
		   op(600, xfx, '==>'), % sequent
		   op(600, xfx, '=clu=>') % cluster sequent
		  ]
	   ).

:-use_module(cluster).


%############    SYNTAX  ############

%
%
%  * Formula A,B ::= atom | ~A | A & B | A v B | A --> B | A <--> B |
%                    box A | diam A | boxp A | diamc A
%
%	where atom is a Prolog atom, boxp and diamc are the new modal
%	connectives.
%
%     Note that:
%
%	  A --> B  ::=   ~A  V B
%        A <--> B  ::=  ( A --> B ) & ( B --> A )
%
%  *  Sets of formulas are represented by Prolog lists.
%
%   * Sequent ::=  Gamma ==> Delta
%
%	 where Gamma and Delta are (possibly empty) sets of  formulas.
%	 Sequents must be in normal form
%
%   * Cluster_sequent ::=   Gamma '=clu=>' Delta
%
%	where Gamma and Delta are (possibly empty) sets of  formulas
%
%  * Proof ::=  proof(Rule_name, End_sequent, Proof_list)
%
%    represents a proof of GS4
%     o	Rule_name is the name of the rule
%     o End_sequent the end sequent
%     o Proof_list is a (possibly empty) list of proofs
%	(the immediate subproofs of Proof)
%
%  * Refutation ::= ref(Rule_name,  End_sequent , Ref_list)
%
%   represents a refutation of RS4.
%
%



%##############    BASIC  DEFINITIONS  ##############

% isempty(+Set)

is_empty([]).


%is_atomic_formula(+Formula)

is_atomic_formula(A) :- atom(A), !.


%is_classical_formula(+Formula)

is_classical_formula( ~ A) :- !,
	is_classical_formula(A).

is_classical_formula( A & B) :- !,
	is_classical_formula(A),
	is_classical_formula(B).

is_classical_formula( A v B) :- !,
	is_classical_formula(A),
	is_classical_formula(B).


is_classical_formula( A --> B) :- !,
	is_classical_formula(A),
	is_classical_formula(B).

is_classical_formula( A <--> B) :- !,
	is_classical_formula(A),
	is_classical_formula(B).


is_classical_formula(A) :- is_atomic_formula(A).




% is_ref(+Term)
%
% true IFF Term represents a refutation



is_ref(Term) :-
	ground(Term), !,
	(
   	 Term =.. [ref | _]  % Term = ref(...)
         ;
	  Term = main_ref(_)
	).


% is_proof(+Term)
%
% true IFF Term represents a proof

is_proof(Term) :-
	ground(Term),
	Term =.. [proof | _].

% is_proof_or_ref(+Term)
%
% true IFF Term represents either a proof or a refutation

is_proof_or_ref(Term) :-   %% NO CUT !!!
	is_ref(Term) ; is_proof(Term).



%########################     PROOF-SEARCH STRATEGY	 ###################################




%  Standard sequents have the form
%
%         Gamma ==> Delta
%
%    where
%
% * Gamma = (TBoxp, TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl)
%
%   with:
%
%    TBoxp        = list of formulas of the kind  boxp E
%    TDiamc	  = list of formulas of the kind  diamc F
%    GBox	  = list of formulas of the kind  box A
%    GBoxDiam     = list of formulas of the kind  box diam B
%    GBoxOrBox    = list of formulas of the kind  box ( C v box D)
%    GBoxOrDiam   = list of formulas of the kind  box ( E v diam F)
%    Gcl	  = list of classical formulas
%
%   A, B, C, D, E, F are classical formulas.
%
%  * Delta =  (DBox, Dcl)
%
%    with:
%
%     DBox =  list of formulas of the kind  box G  (G is a classical formula)
%     Dcl  = list of classical formulas
%
%   Similarly, standard cluster sequents have the form
%
%     Cluster ::=  (GBox,GDiam, Gcl) '=clus=>' Dcl
%
%


gBoxOrBox_to_GBox([],[]).

gBoxOrBox_to_GBox([ box ( C  v box _D) | GBoxOrBox ] , [ box  C | GBox ]) :-
	gBoxOrBox_to_GBox( GBoxOrBox, GBox).




% prove_cluster_seq(+Cluster_sequent, -Result):

prove_cluster_seq( Gamma '=clu=>' Delta, Result):-
	is_ref_cluster_seq( Gamma '=clu=>' Delta, Res), !,
	(
	  Res = true ->
	  Result = ref(refClu,  Gamma '=clu=>' Delta,[] )
	;
	  Result = proof(axClu,  Gamma '=clu=>' Delta,[] )
	).




% seq_to_stSeq(+Sequent, -Standard_Sequent)
%
% translate Sequent in the  standard representation


msg_error_cl(A) :-
    	 write('****** '), write(A), write(' is not a classical formula'),nl.

seq_to_stSeq([boxp Z|Gamma_in] ==> Delta_in,  Gamma_out ==> Delta_out) :- !,
        (
	 is_classical_formula(Z) -> true
	  ;
	   msg_error_cl(Z), abort
	),
	seq_to_stSeq(  Gamma_in ==> Delta_in, Gamma ==> Delta_out),
	Gamma =	(TBoxp ,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
	union(TBoxp, [boxp Z], TBoxp1),
	Gamma_out =  (TBoxp1 ,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl).


seq_to_stSeq([diamc  Z|Gamma_in] ==> Delta_in,  Gamma_out ==> Delta_out) :- !,
	 (
	 is_classical_formula(Z) -> true
	 ;
	 msg_error_cl(Z)
	),
	seq_to_stSeq(  Gamma_in ==> Delta_in, Gamma ==> Delta_out),
	Gamma =	(TBoxp ,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
	union(TDiamc, [diamc Z], TDiamc1),
	Gamma_out =  (TBoxp ,TDiamc1, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl).


seq_to_stSeq([box ( C v box D) |Gamma_in] ==> Delta_in,  Gamma_out ==> Delta_out) :- !,
	 (
	 is_classical_formula(C) -> true
	  ;
	   msg_error_cl(C), abort
	),
	(
	 is_classical_formula(D) -> true
	  ;
	   msg_error_cl(D), abort
	),
	seq_to_stSeq(  Gamma_in ==> Delta_in, Gamma ==> Delta_out),
	Gamma =	(TBoxp ,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
	union(GBoxOrBox, [box (C v box D)], GBoxOrBox1),
	Gamma_out =  (TBoxp ,TDiamc, GBox , GBoxDiam, GBoxOrBox1, GBoxOrDiam ,Gcl).


seq_to_stSeq([box ( E v diam F) |Gamma_in] ==> Delta_in,  Gamma_out ==> Delta_out) :- !,
	(
	 is_classical_formula(E) -> true
	  ;
	   msg_error_cl(E), abort
	),
	(
	 is_classical_formula(F) -> true
	  ;
	   msg_error_cl(F), abort
	),
	seq_to_stSeq(  Gamma_in ==> Delta_in, Gamma ==> Delta_out),
	Gamma =	(TBoxp ,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
	union(GBoxOrDiam, [box (E v diam F)], GBoxOrDiam1),
	Gamma_out =  (TBoxp ,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam1 ,Gcl).

seq_to_stSeq([box diam B |Gamma_in] ==> Delta_in,  Gamma_out ==> Delta_out) :- !,
	(
	 is_classical_formula(B) -> true
	  ;
	   msg_error_cl(B), abort
	),
	seq_to_stSeq(  Gamma_in ==> Delta_in, Gamma ==> Delta_out),
	Gamma =	(TBoxp ,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
	union(GBoxDiam, [box diam B], GBoxDiam1),
	Gamma_out =  (TBoxp ,TDiamc, GBox , GBoxDiam1, GBoxOrBox, GBoxOrDiam ,Gcl).



seq_to_stSeq([box A |Gamma_in] ==> Delta_in,  Gamma_out ==> Delta_out) :- !,
	(
	 is_classical_formula(A) -> true
	  ;
	   msg_error_cl(A), abort
	),
	seq_to_stSeq(  Gamma_in ==> Delta_in, Gamma ==> Delta_out),
	Gamma =	(TBoxp ,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
	union(GBox, [box A], GBox1),
	Gamma_out =  (TBoxp ,TDiamc, GBox1 , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl).


seq_to_stSeq([ Z |Gamma_in] ==> Delta_in,  Gamma_out ==> Delta_out) :- !,
	 (
	 is_classical_formula(Z) -> true
	  ;
	   msg_error_cl(Z), abort
	),
	seq_to_stSeq(  Gamma_in ==> Delta_in, Gamma ==> Delta_out),
	Gamma =	(TBoxp ,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
	union(Gcl, [ Z ], Gcl1),
	Gamma_out =  (TBoxp ,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl1).



seq_to_stSeq([] ==> [box G | Delta_in ], Gamma_out ==> Delta_out) :- !,
	(
	 is_classical_formula(G) -> true
	  ;
	   msg_error_cl(G), abort
	),
	seq_to_stSeq([] ==> Delta_in, Gamma_out ==> Delta),
	Delta = (DBox, Dcl),
	union(DBox, [box G], DBox1),
	Delta_out = (DBox1, Dcl).

seq_to_stSeq([] ==> [ Z | Delta_in ], Gamma_out ==> Delta_out) :- !,
	(
	 is_classical_formula(Z) -> true
	  ;
	   msg_error_cl(Z), abort
	),
	seq_to_stSeq([] ==> Delta_in, Gamma_out ==> Delta),
	Delta = (DBox, Dcl),
	union(Dcl, [ Z ], Dcl1),
	Delta_out = (DBox, Dcl1).


seq_to_stSeq( [] ==> [],  ([],[],[],[],[],[],[])	==>  ([],[])  ).

%  thetaDiam(+Theta, -ThetaDiam)
%
%  ThetaDiam = { diam F | diamc F in Theta }

thetaDiam([ ], []).
thetaDiam([diamc F | Theta ], [ diam F | ThetaDiam ]) :-
	thetaDiam(Theta,ThetaDiam).


%  thetaBox(+Theta, -ThetaBox)
%
%  ThetaBox = { box E | box E in Theta}


thetaBox([ ], []).
thetaBox([ boxp E | Theta ], [ box E | ThetaBox ]) :-
	thetaBox(Theta,ThetaBox).


%  gBoxDiam_to_GDiam(+GBoxDiam, -GDiam}
%
%
%  GDiam = { diam B | box diam B in GBoxDiam }

gBoxDiam_to_GDiam([], []).
gBoxDiam_to_GDiam([box diam B | GBoxDiam ], [ diam B | GDiam  ]) :-
	gBoxDiam_to_GDiam( GBoxDiam ,  GDiam ).

%



%   prove( +Sequent, -Result)
%
%
%  Result is a proof of Input or a refutation of Sequent



prove(Gamma ==> Delta, Result) :-
	seq_to_stSeq( Gamma ==> Delta,  Gamma_st ==> Delta_st),
	prove_st_seq(  Gamma_st  ==> Delta_st, Result).


%   prove( +Standard_sequent, -Result)
%
%
%  Result is a proof of Input or a refutation of Standard_sequent
%  (a sequent in standard form)



prove_st_seq( (TBoxp,TDiamc, GBox , GBoxDiam, [], [] ,Gcl)  ==> ([], Dcl),  Result) :- !,
   apply_box_diam_L( (TBoxp,TDiamc, GBox , GBoxDiam, [], [] ,Gcl)  ==>  ([], Dcl)  ,  Result).



prove_st_seq( (TBoxp,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl)  ==> Delta,  Result) :-
	Gamma = (TBoxp,TDiamc, GBox , GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
	apply_box_or_box_L_main( Gamma ==> Delta,  Refs_boxOrBoxL,  Result_boxOrBoxL), !,
	(
	is_proof_or_ref(Result_boxOrBoxL) ->
	% Result_boxOrBoxL is a proof or a ref of Gamma ==> Delta
	Result = Result_boxOrBoxL
	;
	apply_box_R( Gamma ==> Delta, Refs_boxOrBoxL, Refs_boxR, Result_boxR), !,
	 (
	 is_proof_or_ref(Result_boxR) ->
	 Result = Result_boxR
	 ;
	 apply_box_or_diam_L( Gamma ==> Delta, Refs_boxR, All_Refs,Result_boxOrDiamL),!,
	  (
	  is_proof_or_ref(Result_boxOrDiamL) ->
	  Result = Result_boxOrDiamL
	  ;
	  Result = ref(succ_1, Gamma ==> Delta, All_Refs)
	  )
	 )
	).


%% -----------------------------------------------------------------------

%%  BOX DIAM L


apply_box_diam_L( (ThBoxp,ThDiamc, GBox , GBoxDiam, [], [] ,Gcl) ==>  ([], Dcl), Result) :-
	Gamma =  (ThBoxp,ThDiamc, GBox , GBoxDiam, [], [] ,Gcl),
	Delta =  ([], Dcl),
	thetaDiam(ThDiamc, ThDiam), % ThDiam = { diam F | diamc F in ThDiam}
	gBoxDiam_to_GDiam( GBoxDiam, GDiam), % GDiam = { diam B | box diam B in GBoxDiam }
	union(ThDiam,GDiam,ThDiam_U_GDiam),
	prove_cluster_seq( (GBox,  ThDiam_U_GDiam , Gcl)  '=clu=>' Dcl, Result1 ), !,
	(
	   is_ref(Result1) ->
	     Result = ref(box_diam_l, Gamma ==> Delta,  [Result1]  )
	   ;
	     Result = proof(box_diam_l, Gamma ==> Delta,  [Result1]  )

	).

% ------------------------------------------------------------------------


%%    BOX OR BOX


apply_box_or_box_L_main( Gamma  ==> Delta, Refs_out,Result) :-
  Gamma = (_ThBoxp,_ThDiamc,_GBox,_GBoxDiam, _GBoxOrBox, GBoxOrDiam ,_Gcl),
  Delta = (DBox,_Dcl),
  apply_box_or_box_L_clu( Gamma  ==> Delta, Refs_boxOrBoxL1 ,Result), !,
  (
  is_proof_or_ref(Result) -> true ;
     (
     (is_empty(GBoxOrDiam), is_empty(DBox)) ->
        apply_box_or_box_L( Gamma  ==> Delta, Refs_boxOrBoxL1 ,Result)
     ;
      Refs_out = Refs_boxOrBoxL1
  )
  ).

% check the cluster sequent

apply_box_or_box_L_clu(Gamma  ==> Delta,  Refs_out,Result) :-
  	Gamma =  (ThBoxp,ThDiamc, GBox,GBoxDiam,  GBoxOrBox, GBoxOrDiam ,Gcl),
	Delta = (_DBox,Dcl),
	thetaDiam(ThDiamc,ThDiam),
	gBoxOrBox_to_GBox(GBoxOrBox,GBoxC),
	% GBoxC = { box C | box( C v box D) in GBoxOrBox }
	union(GBox,GBoxC,GBox_U_GBoxC),
	prove_cluster_seq( (GBox_U_GBoxC,  ThDiam ,Gcl)  '=clu=>' Dcl, Result_clu), !,
	(
	  is_ref(Result_clu) ->
	       	Refs_out = [ Result_clu  ]
	 ;
	  % Result_clu is  a proof
	 GBoxOrBox_todo =  GBoxOrBox,
         GBoxOrBox_done = [],
	 Gamma_boxD  = (ThBoxp,ThDiamc,GBox,GBoxDiam, GBoxOrBox_todo,GBoxOrBox_done,GBoxOrDiam,Gcl),
        apply_box_or_box_L_clu_boxD(Gamma_boxD  ==> Delta, [Result_clu],Refs_out,Result)
   ).


apply_box_or_box_L_clu_boxD(
 (ThBoxp,ThDiamc,GBox,GBoxDiam,[ box(C v box D) | GBoxOrBox_todo ],GBoxOrBox_done,GBoxOrDiam ,Gcl) ==> Delta,
		    Proofs_in , Refs_out,Result) :-
	 union(GBoxOrBox_todo, GBoxOrBox_done, GBoxOrBox_todo_U_done),
	 Gamma = (ThBoxp,ThDiamc,GBox,GBoxDiam,[ box(C v box D) |GBoxOrBox_todo_U_done],GBoxOrDiam ,Gcl),
	 union(GBox, [box D], GBox_boxD),
         Gamma_boxD = (ThBoxp,ThDiamc,GBox_boxD,GBoxDiam,GBoxOrBox_todo_U_done,GBoxOrDiam,Gcl),
	 prove_st_seq(Gamma_boxD ==> Delta,  Result_boxD), !,
	 (
	     is_ref(Result_boxD) ->
               Result = ref(box_or_box_l_2, Gamma ==> Delta, [Result_boxD])
	   ;
	    % continue by applying box_or_box_L with the next  box(C v box D) formula
	     Gamma_nextGBoxOrBox = (ThBoxp,ThDiamc, GBox , GBoxDiam, GBoxOrBox_todo, [box( C v box D) | GBoxOrBox_done],
				     GBoxOrDiam ,Gcl),
             apply_box_or_box_L_clu_boxD(  Gamma_nextGBoxOrBox ==> Delta,
				 [Result_boxD | Proofs_in],  Refs_out,Result)
	 ).

% empty GBoxOrBox_todo

apply_box_or_box_L_clu_boxD( (ThBoxp,ThDiamc, GBox , GBoxDiam, [], GBoxOrBox_done,  GBoxOrDiam ,Gcl) ==> Delta,
	 	    Proofs_in , _Refs_out, Result) :-
	GBoxOrBox =  GBoxOrBox_done,
  	Gamma =  (ThBoxp,ThDiamc, GBox , GBoxDiam, GBoxOrBox , GBoxOrDiam ,Gcl),
	Result = proof(box_or_box_l_clu, Gamma ==> Delta,  Proofs_in).


%%% BOX OR BOX L


apply_box_or_box_L(  Gamma  ==> Delta, Refs_in,  Result) :-
	(
	Gamma = ([] , [], _GBox,_GBoxDiam, _GBoxOrBox, _GBoxOrDiam , _Gcl) ->
	   apply_box_or_box_L_emptyTh( Gamma  ==> Delta ,Refs_in, Result)
	 ;
	   apply_box_or_box_L_nonemptyTh( Gamma  ==> Delta ,Refs_in,Result)
	).

% BOX OR BOX L -  empty Theta (the conclusion is a regular sequent)

apply_box_or_box_L_emptyTh( ([],[], GBox , GBoxDiam,[ box(C v box D) |GBoxOrBox1], GBoxOrDiam ,Gcl) ==> Delta,
			 Refs_in,  Result) :-
	 GBoxOrBox = [ box(C v box D) |GBoxOrBox1],
	 Gamma =   ([],[], GBox , GBoxDiam,GBoxOrBox, GBoxOrDiam ,Gcl),
         union(GBox, [ box C], GBox_boxC),
         Gamma_boxC = ([],[], GBox_boxC,GBoxDiam,GBoxOrBox1,GBoxOrDiam ,Gcl),
	 prove_st_seq(Gamma_boxC ==> Delta,  Result_boxC), !,
	 (
	  is_ref(Result_boxC) ->
	     Result = ref(box_or_box_l_1,Gamma ==> Delta,[ Result_boxC])
	     ;
	   %% Result_boxC is a proof
	    union(GBox, [ box D], GBox_boxD),
            Gamma_boxD = ([],[], GBox_boxD,GBoxDiam,GBoxOrBox1,GBoxOrDiam ,[]),
	    prove_st_seq(Gamma_boxD ==> ([],[]),  Result_boxD), !,
	    (
	    is_proof(Result_boxD) ->
	        Result = proof(box_or_box_l_reg,Gamma ==> Delta, [Result_boxC,Result_boxD])
	   ;
	      % Refs_in only contains a ref of a cluster sequent
	      Result = ref(succ_2 , Gamma ==> Delta, [   Result_boxD | Refs_in ] )
	   )
	 ).


% BOX OR BOX L - non empty Theta (the conclusion is a special sequent)


apply_box_or_box_L_nonemptyTh( (ThBoxp,ThDiamc, GBox , GBoxDiam,[ box(C v box D) |GBoxOrBox1], GBoxOrDiam ,Gcl) ==> Delta,
			 Refs_in,  Result) :-
	 GBoxOrBox = [ box(C v box D) |GBoxOrBox1],
	 Gamma =   (ThBoxp,ThDiamc, GBox , GBoxDiam,GBoxOrBox, GBoxOrDiam ,Gcl),
         union(GBox, [ box C], GBox_boxC),
         Gamma_boxC = (ThBoxp,ThDiamc, GBox_boxC,GBoxDiam,GBoxOrBox1,GBoxOrDiam ,Gcl),
	 prove_st_seq(Gamma_boxC ==> Delta,  Result_boxC), !,
	 (
	  is_ref(Result_boxC) ->
	     Result = ref(box_or_box_l_1,Gamma ==> Delta,[ Result_boxC])
	     ;
	   %% Result_boxC is a proof
	    union(GBox, [ box D], GBox_boxD),
            Gamma_boxD = (ThBoxp,ThDiamc, GBox_boxD,GBoxDiam,GBoxOrBox1,GBoxOrDiam ,Gcl),
	    prove_st_seq(Gamma_boxD ==> Delta,  Result_boxD), !,
	    (
	      is_ref(Result_boxD) ->
	         Result = ref(box_or_box_l_2,Gamma ==> Delta,[ Result_boxD])
	     ;
	   %% Result_boxC  and  Result_boxD are proofs
  	    thetaBox(ThBoxp,ThBox),
	    union(ThBox,GBox,ThBox_U_GBox),
	    union(ThBox_U_GBox,[box D], GBox_ThBoxboxD),
	    Gamma_ThBoxboxD = ([],[], GBox_ThBoxboxD,GBoxDiam,GBoxOrBox1,GBoxOrDiam ,[]),
	    prove_st_seq( Gamma_ThBoxboxD ==> ([],[]), Result_ThBoxboxD), !,
	    (
	       is_proof(Result_ThBoxboxD) ->
	        Result = proof(box_or_box_l_spec,Gamma ==> Delta, [Result_boxC,Result_boxD, Result_ThBoxboxD])
	   ;
	     % Refs_in only contains a ref of a cluster sequent
	      Result = ref(succ_2 , Gamma ==> Delta, [   Result_ThBoxboxD | Refs_in ] )
	   )
	  )
	 ).

%--------------------------------------------------------------

%%%  BOX R


apply_box_R( Gamma  ==> Delta ,Refs_in, Refs_out,Result) :-
   (
     Gamma = ([] , [], _GBox,_GBoxDiam, _GBoxOrBox, _GBoxOrDiam , _Gcl) ->
       apply_box_R_emptyTh( Gamma  ==> Delta ,Refs_in, Refs_out,Result)
      ;
       apply_box_R_nonemptyTh( Gamma  ==> Delta ,Refs_in, Refs_out,Result)
   ).


%%%  BOX R -  empty Theta (the conclusion is a regular sequent)

apply_box_R_emptyTh( Gamma  ==> (DBox,Dcl) ,Refs_in, Refs_out,Result) :-
  DBox_todo = DBox,
  DBox_done = [],
  apply_box_R_G_emptyTh( Gamma  ==> ( DBox_todo,DBox_done,Dcl), Refs_in,Refs_out ,Result).


apply_box_R_G_emptyTh( Gamma  ==> ( [box G | DBox_todo], DBox_done,Dcl), Refs_in,Refs_out, Result) :-
   Gamma = ([],[],GBox,GBoxDiam, GBoxOrBox, GBoxOrDiam ,_Gcl),
   union(DBox_todo,DBox_done,DBox1),
   Delta = ([box G | DBox1], Dcl),
   Gamma_G = ([],[], GBox,GBoxDiam, GBoxOrBox, GBoxOrDiam ,[]),
    prove_st_seq(Gamma_G ==> ([],[G]), Result_G), !,
    (
      is_proof(Result_G) ->
         Result = proof(box_r_reg, Gamma ==> Delta, [ Result_G])
      ;
     % continue by applying box_or_box_L with the next  GBoxOrBox formula
     apply_box_R_G_emptyTh(  Gamma  ==> ( DBox_todo, [box G | DBox_done],Dcl), [Result_G| Refs_in],Refs_out, Result)
   ).

% empty DBox_todo
apply_box_R_G_emptyTh( _Gamma  ==> ( [], _DBox_done,_Dcl), Refs_in, Refs_out, _Result) :-
	Refs_out = Refs_in.



%%%  BOX R -  non empty  Theta (the conclusion is a special sequent)

apply_box_R_nonemptyTh( Gamma  ==> (DBox,Dcl) ,Refs_in, Refs_out,Result) :-
  DBox_todo = DBox,
  DBox_done = [],
  apply_box_R_G_nonemptyTh( Gamma  ==> ( DBox_todo,DBox_done,Dcl), Refs_in,Refs_out ,Result).


apply_box_R_G_nonemptyTh( Gamma  ==> ( [box G | DBox_todo], DBox_done,Dcl), Refs_in,Refs_out, Result) :-
   Gamma = (ThBoxp,ThDiamc,GBox,GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
   union(DBox_todo,DBox_done,DBox1),
   Delta = ([box G | DBox1], Dcl),
   union(ThDiamc, [diamc ~G ], ThDiamc_negG ),
   Gamma_diamcNegG = (ThBoxp,ThDiamc_negG,GBox,GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
   prove_st_seq(Gamma_diamcNegG ==> (DBox1,Dcl),Result_diamcNegG), !,
   (
     is_ref(Result_diamcNegG) ->
       Result = ref(box_r, Gamma ==> Delta, [Result_diamcNegG])
   ;
     % Result_diamcNegG is a proof
    thetaBox(ThBoxp,ThBox),
    union(ThBox,GBox,ThBox_U_GBox),
    Gamma_G = ([],[],ThBox_U_GBox,GBoxDiam, GBoxOrBox, GBoxOrDiam ,[]),
    prove_st_seq(Gamma_G ==> ([],[G]), Result_G), !,
    (
      is_proof(Result_G) ->
         Result = proof(box_r_spec, Gamma ==> Delta, [ Result_diamcNegG, Result_G])
      ;
     % continue by applying box_or_box_L with the next  GBoxOrBox formula
     apply_box_R_G_nonemptyTh(  Gamma  ==> ( DBox_todo, [box G | DBox_done],Dcl), [Result_G| Refs_in],Refs_out, Result)
    )
   ).

% empty DBox_todo
apply_box_R_G_nonemptyTh( _Gamma  ==> ( [], _DBox_done,_Dcl), Refs_in,Refs_out, _Result) :-
	Refs_out = Refs_in.

%

%%	BOX OR DIAM L

% ev_true_ref(+F, +Ref)
%
%  F is eventually true in refutation Ref
% ref(refClu, ([box~b1], [], [f1])'=clu=>'[g2], []))

%%  We do not consider cluster sequents


% ev_true_ref(F,    ref(refClu,  (_GBox,_GDiam,Gcl)  '=clu=>' _Delta ,[] ) ) :-
%  member(F , Gcl).




ev_true_ref(F, ref(_Rule,(_ThBoxp, _ThDiamc, GBox, GBoxDiam, _GBoxOrBox, _GBoxOrDiam ,Gcl) ==> _Delta , Ref_List)) :-
   member(F , Gcl) , !
   ;
   member( box F , GBox), !
   ;
   member(box diam F, GBoxDiam), !
   ;
   ev_true_ref_list(F, Ref_List).



% ev_true_ref(+F, +Ref_List)
%
%  F is eventually true in the set of refutations Ref_List

ev_true_ref_list(F, [Ref | Ref_List]) :-
	ev_true_ref(F, Ref), ! ;
	ev_true_ref_list(F, Ref_List).



% -----------------------------------------------------------------------

%%	BOX OR DIAM L

apply_box_or_diam_L( Gamma  ==> Delta, Refs_in,Refs_out,Result) :-
  (
     Gamma = ([],[],_GBox,_GBoxDiam, _GBoxOrBox, _GBoxOrDiam ,_Gcl) ->
	apply_box_or_diam_L_emptyTh( Gamma  ==> Delta, Refs_in,Refs_out,Result)
  ;
  	apply_box_or_diam_L_nonemptyTh( Gamma  ==> Delta, Refs_in,Refs_out,Result)
   ).


%%  BOX OR DIAM L - empty Theta (the conclusion is a regular sequent)

apply_box_or_diam_L_emptyTh( Gamma  ==> Delta, Refs_in,Refs_out,Result) :-
  Gamma = ([],[],GBox,GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
  GBoxOrDiam_todo = GBoxOrDiam,
  GBoxOrDiam_done = [],
  Gamma_box_or_diam_L=  ([],[],GBox,GBoxDiam, GBoxOrBox, GBoxOrDiam_todo,   GBoxOrDiam_done,Gcl),
  apply_box_or_diam_L_F_emptyTh( Gamma_box_or_diam_L  ==> Delta, Refs_in,Refs_out,Result).



 apply_box_or_diam_L_F_emptyTh(
  ([],[],GBox,GBoxDiam, GBoxOrBox, [ box(E v diam F) |GBoxOrDiam_todo], GBoxOrDiam_done,Gcl)  ==> Delta,
			Refs_in,Refs_out,Result) :-
      (
 	  ev_true_ref_list(F, Refs_in) ->
 	  % continue by applying box_or_box_L with the next  box(E v diam F) formula
         Gamma_nextGBoxOrDiam = ([],[],GBox,GBoxDiam, GBoxOrBox,  GBoxOrDiam_todo,
				 [ box(E v diam F) | GBoxOrDiam_done],Gcl),
	  apply_box_or_diam_L_F_emptyTh( Gamma_nextGBoxOrDiam ==> Delta, Refs_in, Refs_out, Result) ,!
	 ;
	union( GBoxOrDiam_todo, GBoxOrDiam_done, GBoxOrDiam_todo_U_done),
	Gamma =  ([],[], GBox , GBoxDiam, GBoxOrBox,   [ box(E v diam F) |GBoxOrDiam_todo_U_done],Gcl),
        union(GBox, [ box E], GBox_boxE),
        Gamma_boxE =  ([],[],GBox_boxE,GBoxDiam, GBoxOrBox,  GBoxOrDiam_todo_U_done,Gcl),
	prove_st_seq(Gamma_boxE ==> Delta,  Result_boxE), !,
	(
	   is_ref(Result_boxE) ->
	      Result = ref(box_or_diam_l_1, Gamma ==> Delta, [Result_boxE])
	;
	    %% Result_box E is a proof
	    Gamma_boxpE_F =  ([boxp E],[],GBox,GBoxDiam, GBoxOrBox, GBoxOrDiam_todo_U_done,[F]),
	    prove_st_seq(  Gamma_boxpE_F ==> ([],[]), Result_boxpE_F), !,
            (
	      is_proof( Result_boxpE_F) ->
	          Result = proof(box_or_diam_l_reg, Gamma ==> Delta, [Result_boxE, Result_boxpE_F])
	    ;
	    % continue by applying box_or_box_L with the next box(E v diam F) formula
            Gamma_nextGBoxOrDiam = ([],[],GBox,GBoxDiam, GBoxOrBox,  GBoxOrDiam_todo,
				       [ box(E v diam F) | GBoxOrDiam_done],Gcl),
	    apply_box_or_diam_L_F_emptyTh( Gamma_nextGBoxOrDiam ==> Delta, [Result_boxpE_F |Refs_in], Refs_out, Result)
	    )
	)
	).


% empty GBoxOrDiam_todo

apply_box_or_diam_L_F_emptyTh(([],[],_GBox,_GBoxDiam, _GBoxOrBox, [], _GBoxOrDiam_done, _Gcl)  ==> _Delta,
			      Refs_in,Refs_out,_Result) :-
           	Refs_out = Refs_in.


%%	BOX OR DIAM L - non empty Theta (the conclusion is a special sequent)

apply_box_or_diam_L_nonemptyTh( Gamma  ==> Delta, Refs_in,Refs_out,Result) :-
  Gamma = (ThBoxp,ThDiamc,GBox,GBoxDiam, GBoxOrBox, GBoxOrDiam ,Gcl),
  GBoxOrDiam_todo = GBoxOrDiam,
  GBoxOrDiam_done = [],
  Gamma_box_or_diam_L=  (ThBoxp,ThDiamc,GBox,GBoxDiam, GBoxOrBox, GBoxOrDiam_todo,   GBoxOrDiam_done,Gcl),
  apply_box_or_diam_L_F_nonemptyTh( Gamma_box_or_diam_L  ==> Delta, Refs_in,Refs_out,Result).



 apply_box_or_diam_L_F_nonemptyTh(
   (ThBoxp,ThDiamc,GBox,GBoxDiam, GBoxOrBox, [ box(E v diam F) |GBoxOrDiam_todo], GBoxOrDiam_done,Gcl)  ==> Delta,
			   Refs_in,Refs_out,Result) :-
      (
	  ev_true_ref_list(F, Refs_in) ->
	Gamma_nextGBoxOrDiam = (ThBoxp,ThDiamc,GBox,GBoxDiam, GBoxOrBox,  GBoxOrDiam_todo,
				[ box(E v diam F) | GBoxOrDiam_done],Gcl),
	  apply_box_or_diam_L_F_nonemptyTh( Gamma_nextGBoxOrDiam ==> Delta, Refs_in, Refs_out, Result) ,!
	 ;
	union( GBoxOrDiam_todo, GBoxOrDiam_done, GBoxOrDiam_todo_U_done),
	Gamma =  (ThBoxp,ThDiamc, GBox , GBoxDiam, GBoxOrBox,   [ box(E v diam F) |GBoxOrDiam_todo_U_done],Gcl),
        union(GBox, [ box E], GBox_boxE),
        Gamma_boxE =  (ThBoxp,ThDiamc,GBox_boxE,GBoxDiam, GBoxOrBox,  GBoxOrDiam_todo_U_done,Gcl),
	prove_st_seq(Gamma_boxE ==> Delta,  Result_boxE), !,
	(
	   is_ref(Result_boxE) ->
	      Result = ref(box_or_diam_l_1, Gamma ==> Delta, [Result_boxE])
	;
	    %% Result_box E is a proof
           union(ThDiamc, [diamc F], ThDiamc_diamcF),
	   union(ThBoxp, [boxp E], ThBoxp_boxpE),
	   Gamma_boxpE_diamcF = (ThBoxp_boxpE,ThDiamc_diamcF,GBox_boxE,GBoxDiam, GBoxOrBox, GBoxOrDiam_todo_U_done,Gcl),
	   prove_st_seq(Gamma_boxpE_diamcF ==> Delta,  Result_boxpE_diamcF), !,
           (
	      is_ref(Result_boxpE_diamcF) ->
	         Result = ref(box_or_diam_l_2, Gamma ==> Delta, [Result_boxpE_diamcF ])
	   ;
	    %% Result_boxpEdiamcF is a proof
	    thetaBox(ThBoxp,ThBox),
	    union(GBox,ThBox,GBox_U_ThBox),
	    Gamma_boxpE_F =  ([boxp E],[],GBox_U_ThBox,GBoxDiam, GBoxOrBox, GBoxOrDiam_todo_U_done,[F]),
	    prove_st_seq(  Gamma_boxpE_F ==> ([],[]), Result_boxpE_F), !,
            (
               is_proof( Result_boxpE_F) ->
	          Result = proof(box_or_diam_l_spec, Gamma ==> Delta, [Result_boxE,Result_boxpE_diamcF,	Result_boxpE_F])
	    ;
	    %%
	    % continue by applying box_or_box_L with the next  box(E v diam F) formula
             Gamma_nextGBoxOrDiam = (ThBoxp,ThDiamc,GBox,GBoxDiam, GBoxOrBox,  GBoxOrDiam_todo,
				     [ box(E v diam F) | GBoxOrDiam_done],Gcl),
	     apply_box_or_diam_L_F_nonemptyTh( Gamma_nextGBoxOrDiam ==> Delta, [Result_boxpE_F |Refs_in], Refs_out, Result)
	    )
	   )
	)
	).


% empty GBoxOrDiam_todo

apply_box_or_diam_L_F_nonemptyTh((_ThBoxp,_ThDiamc,_GBox,_GBoxDiam, _GBoxOrBox, [], _GBoxOrDiam_done, _Gcl)  ==> _Delta,
				 Refs_in,Refs_out,_Result) :-
           	Refs_out = Refs_in.










































